/* LBfixed.c  */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include <signal.h>
#include <sys/wait.h>
#include <pthread.h>
#include <fcntl.h>
#include <sys/ioctl.h>

#include "Util.h"
#include "VoIPpkt.h"

/* #define VICDEBUG */

#define P(X) do { fprintf(stderr,X "\n"); fflush(stderr); } while(0)

typedef struct {
	uint16_t port_number;
	int attivo;
} UDPdest;
UDPdest udpdest;

fd_set rdset, all;
int maxfd;
int numspediti=0;
int numscartati=0;
int printed=0;

void sig_print(int signo)
{
	if(printed==0)
	{
		printed=1;
		if(signo==SIGINT)		printf("SIGINT\n");
		else if(signo==SIGHUP)	printf("SIGHUP\n");
		else if(signo==SIGTERM)	printf("SIGTERM\n");
		else					printf("other signo\n");
		printf("\n");
		/*
		for(i=0;i<MAXNUMCONNECTIONS;i++)
		{
			printf("canale %d: sent %d received %d  tot %d\n",
					i, Ricevuti[i][0],Ricevuti[i][1], Ricevuti[i][0]+Ricevuti[i][1] );
		}
		*/
		printf("numspediti %d  numscartati %d\n", numspediti, numscartati );
		fflush(stdout);
	}
	exit(0);
	return;
}

void stampa_fd_set(char *str, fd_set *pset)
{
	int i;
	printf("%s ",str);
	for(i=0;i<100;i++) if(FD_ISSET(i,pset)) printf("%d ", i);
;	printf("\n");
}

int send_udp(uint32_t socketfd, char *buf, uint32_t len, uint16_t port_number_local, char *IPaddr, uint16_t port_number_dest)
{
	int ris;
	struct sockaddr_in To;
	int addr_size;

	/* assign our destination address */
	memset( (char*)&To,0,sizeof(To));
	To.sin_family		=	AF_INET;
	To.sin_addr.s_addr  =	inet_addr(IPaddr);
	To.sin_port			=	htons(port_number_dest);

	addr_size = sizeof(struct sockaddr_in);
	/* send to the address */
	ris = sendto(socketfd, buf, len , MSG_NOSIGNAL, (struct sockaddr*)&To, addr_size);
	if (ris < 0) {
		printf ("sendto() failed, Error: %d \"%s\"\n", errno,strerror(errno));
		return(0);
	}
	return(ris);
}

long int compute_delay(uint32_t *buf)
{
	struct timeval sent, now, tempoimpiegato;
	uint32_t idmsg;
	long int msec;

	idmsg=buf[0];
	memcpy( (char*)&sent, (char*)&(buf[1]), sizeof(struct timeval) );
	gettimeofday(&now,NULL);
	/*
	fprintf( f,	"\nsent: %lu : %ld sec %ld usec\n", idmsg, sent.tv_sec, sent.tv_usec );
	fprintf( f,	"now:  %lu : %ld sec %ld usec\n", idmsg, now.tv_sec, now.tv_usec );
	*/
	tempoimpiegato=differenza(now,sent);
	msec=(tempoimpiegato.tv_sec*1000)+(tempoimpiegato.tv_usec/1000);
	/*
	fprintf(	f, "tempoimpiegato: %lu : %ld sec %ld usec\n\n", 
			idmsg, tempoimpiegato.tv_sec, tempoimpiegato.tv_usec );
	*/
	return(msec);
}

#define PARAMETRIDEFAULT "./LBfixed.exe 11001 10001"
void usage(void) 
{  printf ( "usage:  ./LBfixed.exe LOCALPORTFIXED LOCALPORTDATA\n"
			"esempio: "  PARAMETRIDEFAULT "\n");
}

int main(int argc, char *argv[])
{
	uint16_t localportfromfixed, localportdatamonitor;
	struct sockaddr_in Cli;
	unsigned int len;
	int32_t monitordatafd;
	int ris, listening_fixedfd, appfixedfd;

	printf ("uso i parametri di default \n%s\n", PARAMETRIDEFAULT );
	localportfromfixed = 11001;
	localportdatamonitor = 10001;

	udpdest.attivo=0;

	init_random();

	ris=UDP_setup_socket_bound( &monitordatafd, localportdatamonitor, 65535, 65535 );
	if (!ris) {	printf ("UDP_setup_socket_bound() failed\n"); exit(1); }

	FD_ZERO(&all);
	FD_SET(monitordatafd,&all);
	if(monitordatafd>maxfd)
		maxfd=monitordatafd;

	appfixedfd=0;
	/* attendo connessione dalla applicazione lato fixed */
	ris=TCP_setup_socket_listening( &listening_fixedfd, localportfromfixed, 300000, 300000, 1);
	if (!ris)
	{	printf ("TCP_setup_socket_listening() failed\n");
		exit(1);
	}
	do { memset (&Cli, 0, sizeof (Cli));		
		len = sizeof (Cli);
		appfixedfd = accept ( listening_fixedfd, (struct sockaddr *) &Cli, &len);
	} while ( (appfixedfd<0) && (errno==EINTR) );
	if (appfixedfd < 0 ) {	
		perror("accept() failed: \n");
		exit (1);
	}
	/* NO NAGLE per i tcp */
	ris=SetsockoptTCPNODELAY(appfixedfd,1); if (!ris) { fprintf(stderr,"unable to setup TCPNODELAY option\n"); exit(5); }

	/* chiusura del socket TCP listening e setup dei socket UDP	*/
	close(listening_fixedfd);
	FD_SET(appfixedfd,&all);
	if(appfixedfd>maxfd)
		maxfd=appfixedfd;

	for(;;)
	{
		do {
			rdset=all;

#ifdef VICDEBUG
			stampa_fd_set("rdset prima",&rdset);
#endif

			ris=select(maxfd+1,&rdset,NULL,NULL,NULL);
		} while( (ris<0) && (errno==EINTR) );
		if(ris<0) {
			perror("select failed: ");
			exit(1);
		}

#ifdef VICDEBUG
		stampa_fd_set("rdset dopo",&rdset);
#endif

		if( FD_ISSET(appfixedfd,&rdset) )
		{
			uint32_t buf[PKTSIZE];
			uint32_t idmsg;

			/* ricevo un pkt */
			ris=Readn(appfixedfd,(char*)buf,PKTSIZE);
			if(ris!=PKTSIZE) { fprintf(stderr,"recv from appfixed failed, received %d: ", ris); exit(9); }

			/* spedisco i pkt */
			if(udpdest.attivo==1)  {
				idmsg=buf[0];
				ris=send_udp(monitordatafd, (char*)buf, PKTSIZE , 1000, "127.0.0.1", udpdest.port_number );
				if(ris)  {
					;
#ifdef OUTPUT_MEDIO
					fprintf(stderr,"pkt id %u sent to %d\n", idmsg, udpdest.port_number);
#endif
				}
				else {
					fprintf(stderr,"pkt id %u NOT sent\n", idmsg );
				}
			}
		}

		/* se arriva qualche pacchetto dati UDP dal monitor, leggo!!!! */
		if(FD_ISSET(monitordatafd,&rdset))  {
			uint32_t buf[ (65536/sizeof(uint32_t)) ];
			struct sockaddr_in From;
			unsigned int Fromlen;
			int ris;

			memset(&From,0,sizeof(From));
			Fromlen=sizeof(struct sockaddr_in);
			ris = recvfrom ( monitordatafd, (char*)buf, (int)65536, 0, (struct sockaddr*)&From, &Fromlen);
			if (ris<0) {
				if(errno!=EINTR) {	
					fprintf(stderr,"recvfrom() failed, Error: %d \"%s\"\n", errno,strerror(errno));
					/*
					close(monitordatafd);
					exit(8);
					*/
					fprintf(stderr,"ma non ho chiuso  il socket");
				}
			}
			else {
				int ris2;
				long int msecdelay;
				uint32_t idletto;
				
				idletto=buf[0];
				msecdelay=compute_delay(buf);
#ifdef OUTPUT_MEDIO
				fprintf(stderr,"LBfixed pkt received size %d id %u delay %ld msec \n", ris, idletto, msecdelay);
#endif
				/*	seleziono come porta del monitor a cui destinare i prossimi pacchetti
					la porta dati UDP del monitor da cui ho ricevuto questo ultimo pacchetto
				*/
				udpdest.attivo=1;
				udpdest.port_number=ntohs(From.sin_port);
				/* spedisco all'applicazione */
				ris2=Sendn(appfixedfd,(char*)buf,ris);
				if(ris2!=ris) { fprintf(stderr,"recv from appfixedfd failed, received %d: ", ris); exit(9); }
			}
		}

	} /* fine for ;; */
	return(0);
}

